/*
    uhttp_client.c - Minimalistic HTTP client library.
    Copyright (C) 2008  Avelino Herrera Morales

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include "uhttp_client.h"

#define  CHUNK_SIZE  1024

static struct hostent *he;
static struct sockaddr_in remote;
static int sockfd;

#define  RECV_BUFFER_SIZE  1024

static unsigned char header_buffer[RECV_BUFFER_SIZE];
static unsigned char *header_buffer_p;

/*
#define  READ_BUFFER_SIZE  1024

unsigned char read_buffer[READ_BUFFER_SIZE];
int readed;
*/
static char parameter_name[256];
static char parameter_value[256];
int uhttp_status = UHTTP_STATUS_DISCONNECTED;

int uhttp_errno = 0;
const char *uhttp_errlist[] = {
	"Ok",
	"Already connected",
	"URL parse error",
	"Status not idle",
	"Unable to resolve hostname",
	"Socket error",
	"Unable to connect to remote host",
	"Unable to send data",
	"Unable to receive data",
	"Try again",
	"Out of memory",
	"Unable to reuse address"
};

static char complete_url[1024];
static char host[80];
static char path[80];
static int port;
static unsigned char *to_send_buffer = NULL;
static unsigned char *to_send_buffer_p;
static unsigned char aux_buffer[1024];
static int to_send_buffer_size;
static int to_send_buffer_i;
static int reading_header;
static unsigned char **to_receive_buffer;
static unsigned char *to_receive_buffer_p;
static int *to_receive_buffer_size;
static int to_receive_buffer_i;
static char cookie[2048];
static int is_filed_request = 0;
static int is_filep_request = 0;
static int is_file_request = 0;
static int filed;
static FILE *filep;
static int file_size;
static int remain_file_size;
static unsigned char uhttp_data[CHUNK_SIZE];

static int use_proxy = 0;
static char uhttp_proxy_host[1024] = "";
static int uhttp_proxy_port;
static char uhttp_proxy_user[80] = "";
static char uhttp_proxy_passwd[80] = "";
static char additional_header[1024] = "";
void (*uhttp_sending_callback)(int) = NULL;
void (*uhttp_receiving_callback)(int) = NULL;

static int uhttp_transfer_speed_enable = 0;
unsigned long long uhttp_send_transfer_speed = 0;  // kbytes/seg
unsigned long long uhttp_recv_transfer_speed = 0;  // kbytes/seg
static unsigned long long uhttp_transfer_speed = 0;       // kbytes/seg
static int uhttp_transfered = 0;
static unsigned long long uhttp_transfer_initial_timestamp;


static unsigned long long get_uhttp_timestamp(struct timeval *tv) {
	return ((unsigned long long) tv->tv_usec / 1000ULL) + ((unsigned long long) tv->tv_sec * 1000ULL);
}


void uhttp_enable_transfer_speed_measurement(int enable) {
	uhttp_transfer_speed_enable = enable;
}


static void uhttp_transfer_speed_init(void) {
	struct timeval tv;

	if (!uhttp_transfer_speed_enable)
		return;
	gettimeofday(&tv, NULL);
	uhttp_transfer_initial_timestamp = get_uhttp_timestamp(&tv);
	//printf("initial ts = %llu\n", uhttp_transfer_initial_timestamp);
	uhttp_transfered = 0;
	uhttp_transfer_speed = 0;
}


static void uhttp_transfer_speed_work(int amount) {
	struct timeval tv;
	unsigned long long bytes_msec, diff, aux;

	if (!uhttp_transfer_speed_enable)
		return;
	gettimeofday(&tv, NULL);
	aux = get_uhttp_timestamp(&tv);
	//printf("new ts = %llu\n", aux);
	diff = aux - uhttp_transfer_initial_timestamp;
	uhttp_transfered += amount;
	if (diff > 0) {
		bytes_msec = uhttp_transfered / diff;
		uhttp_transfer_speed = (bytes_msec * 1000ULL) / 1024ULL;
	}
}


#define BASE64_LENGTH(len) (4 * (((len) + 2) / 3))


static void base64_encode (const char *s, char *store, int length) {
	// conversion table
	static char tbl[64] = {
		'A','B','C','D','E','F','G','H',
		'I','J','K','L','M','N','O','P',
		'Q','R','S','T','U','V','W','X',
		'Y','Z','a','b','c','d','e','f',
		'g','h','i','j','k','l','m','n',
		'o','p','q','r','s','t','u','v',
		'w','x','y','z','0','1','2','3',
		'4','5','6','7','8','9','+','/'
	};
	int i;
	unsigned char *p = (unsigned char *) store;

	// transform the 3x8 bits to 4x6 bits, as required by base64
	for (i = 0; i < length; i += 3) {
		*p++ = tbl[s[0] >> 2];
		*p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
		*p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
		*p++ = tbl[s[2] & 0x3f];
		s += 3;
	}
	// pad the result if necessary
	if (i == length + 1)
		*(p - 1) = '=';
	else if (i == length + 2)
		*(p - 1) = *(p - 2) = '=';
	*p = '\0';
}


static int parse_url(char *url, char *host, int *port, char *path) {
	char temp[80];
	strcpy(host, "");
	*port = 80;
	strcpy(path, "/");
	if (sscanf(url, "http://%[^/]%s", temp, path) < 1)
		return 1;
	sscanf(temp, "%[^:]:%d", host, port);
	uhttp_errno = 0;
	return 0;
}


int uhttp_get_socket_fd(void) {
	return sockfd;
}


void uhttp_set_proxy_configuration(char *user, char *passwd, char *host, int port) {
	additional_header[0] = 0;
	if ((host == NULL) || (*host == 0))
		use_proxy = 0;
	else {
		use_proxy = 1;
		if (uhttp_proxy_host != host)
			strcpy(uhttp_proxy_host, host);
		uhttp_proxy_port = port;
		if ((user != NULL) && (passwd != NULL) && (*user != 0) && (*passwd != 0)) {
			int aux_len;
			if (uhttp_proxy_user != user)
				strcpy(uhttp_proxy_user, user);
			if (uhttp_proxy_passwd != passwd)
				strcpy(uhttp_proxy_passwd, passwd);
			strcpy(additional_header, "Proxy-Authorization: Basic ");
			aux_len = strlen(additional_header);
			sprintf((char *) aux_buffer, "%s:%s", user, passwd);
			base64_encode((char *) aux_buffer, &(additional_header[aux_len]), strlen((char *) aux_buffer));
			strcat(additional_header, "\r\n");
		}
		else
			*uhttp_proxy_user = *uhttp_proxy_passwd = 0;
	}
}


int uhttp_get_proxy_configuration(char *user, char *passwd, char *host, int *port) {
	if (!use_proxy)
		return 1;
	strcpy(user, uhttp_proxy_user);
	strcpy(passwd, uhttp_proxy_passwd);
	strcpy(host, uhttp_proxy_host);
	*port = uhttp_proxy_port;
	return 0;
}


void uhttp_set_proxy_configuration_from_environment(void) {
	char proxy_port_str[80];
	char *p;

	*uhttp_proxy_host = 0;
	p = getenv("HTTP_PROXY_HOST");
	if ((p != NULL) && (*p != 0))
		strcpy(uhttp_proxy_host, p);
	*proxy_port_str = 0;
	uhttp_proxy_port = 0;
	p = getenv("HTTP_PROXY_PORT");
	if ((p != NULL) && (*p != 0)) {
		strcpy(proxy_port_str, p);
		uhttp_proxy_port = atoi(proxy_port_str);
	}
	*uhttp_proxy_user = 0;
	p = getenv("HTTP_PROXY_USER");
	if ((p != NULL) && (*p != 0))
		strcpy(uhttp_proxy_user, p);
	*uhttp_proxy_passwd = 0;
	p = getenv("HTTP_PROXY_PASSWD");
	if ((p != NULL) && (*p != 0))
		strcpy(uhttp_proxy_passwd, p);
	if ((uhttp_proxy_port > 0) && (*uhttp_proxy_host != 0)) {
		if ((*uhttp_proxy_user != 0) && (*uhttp_proxy_passwd != 0))
			uhttp_set_proxy_configuration(uhttp_proxy_user, uhttp_proxy_passwd, uhttp_proxy_host, uhttp_proxy_port);
		else
			uhttp_set_proxy_configuration(NULL, NULL, uhttp_proxy_host, uhttp_proxy_port);
	}
}


/*
 * errors: UHTTP_ERROR_ALREADY_CONNECTED UHTTP_ERROR_PARSE_URL
 */
int uhttp_connect(char *url) {
	uhttp_send_transfer_speed = 0;
	uhttp_recv_transfer_speed = 0;
	uhttp_errno = 0;
	is_file_request = 0;
	strcpy(complete_url, url);
	if (uhttp_status != UHTTP_STATUS_DISCONNECTED) {
		uhttp_errno = UHTTP_ERROR_ALREADY_CONNECTED;
		return 1;
	}
	if (use_proxy) {
		strcpy(host, uhttp_proxy_host);
		port = uhttp_proxy_port;
		strcpy(path, url);
	}
	else {
		if (parse_url(url, host, &port, path) != 0) {
			uhttp_status = UHTTP_STATUS_ERROR;
			uhttp_errno = UHTTP_ERROR_PARSE_URL;
			return 1;
		}
	}
	/*
	printf("host = '%s'\n", host);
	printf("port = %d\n", port);
	printf("path = '%s'\n", path);
	*/
	cookie[0] = 0;
	uhttp_status = UHTTP_STATUS_IDLE;
	uhttp_errno = 0;
	return 0;
}


void uhttp_disconnect(void) {
	if ((to_send_buffer != aux_buffer) && (to_send_buffer != NULL)) {
		free(to_send_buffer);
		to_send_buffer = NULL;
	}
	to_send_buffer_size = 0;
	close(sockfd);
	sockfd = -1;
	uhttp_status = UHTTP_STATUS_DISCONNECTED;
	uhttp_errno = 0;
}


/*
 * errors: UHTTP_ERROR_NOT_IDLE UHTTP_ERROR_UNABLE_TO_RESOLVE UHTTP_ERROR_SOCKET UHTTP_ERROR_NO_MEM UHTTP_ERROR_UNABLE_TO_REUSE_ADDR
 */
int uhttp_prepare_request(unsigned char *to_send, int to_send_size, unsigned char **to_receive, int *to_receive_size) {
	unsigned char *p;
	int ret, on;

	uhttp_errno = 0;
	if (uhttp_status != UHTTP_STATUS_IDLE) {
		uhttp_errno = UHTTP_ERROR_NOT_IDLE;
		return 1;
	}
	uhttp_send_transfer_speed = 0;
	uhttp_recv_transfer_speed = 0;
	to_receive_buffer = to_receive;
	to_receive_buffer_size = to_receive_size;

	/* Puesto por Marcos al final de la funcin
	uhttp_status = UHTTP_STATUS_CONNECTING;
	*/

	/* prepare the send buffer */
	/* generate http header */
	if (cookie[0] != 0) {
		if (to_send_size > 0)
			sprintf((char *) aux_buffer, "POST %s HTTP/1.1\r\nHost: %s\r\nCookie: %s\r\nContent-Type: application/octet-stream\r\nContent-Length: %d\r\n%s\r\n", path, host, cookie, to_send_size, additional_header);
		else
			sprintf((char *) aux_buffer, "GET %s HTTP/1.1\r\nHost: %s\r\nCookie: %s\r\n%s\r\n", path, host, cookie, additional_header);
		//printf("cookie = '%s'\n", cookie);
	}
	else {
		if (to_send_size > 0)
			sprintf((char *) aux_buffer, "POST %s HTTP/1.1\r\nHost: %s\r\nContent-Type: application/octet-stream\r\nContent-Length: %d\r\n%s\r\n", path, host, to_send_size, additional_header);
		else
			sprintf((char *) aux_buffer, "GET %s HTTP/1.1\r\nHost: %s\r\n%s\r\n", path, host, additional_header);
	}
	if (to_send_size > 0) {
		/* alloc space to entiry request (header + data) */
		to_send_buffer_size = strlen((char *) aux_buffer) + to_send_size;
		if ((to_send_buffer = (unsigned char *) malloc(to_send_buffer_size)) == NULL) {
			uhttp_errno = UHTTP_ERROR_NO_MEM;
			return 1;
		}
		/* copy the http header to the buffer */
		*to_send_buffer = 0;
		strcpy((char *) to_send_buffer, (char *) aux_buffer);
		/* concatenate the data after the http header */
		p = to_send_buffer + strlen((char *) aux_buffer);
		memcpy(p, to_send, to_send_size);
	}
	else {
		to_send_buffer = (unsigned char *) aux_buffer;
		to_send_buffer_size = strlen((char *) aux_buffer);
	}
	/* init pointers */
	to_send_buffer_p = to_send_buffer;
	to_send_buffer_i = 0;
	if ((he = gethostbyname(host)) == NULL) {
		uhttp_errno = UHTTP_ERROR_UNABLE_TO_RESOLVE;
		return 1;
	}
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		uhttp_errno = UHTTP_ERROR_SOCKET;
		return 1;
	}
	on = 1;
	if ((ret = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) == -1) {
		uhttp_errno = UHTTP_ERROR_UNABLE_TO_REUSE_ADDR;
		return 1;
	}
	/* Puesto por Marcos al final de la funcin */
	uhttp_status = UHTTP_STATUS_CONNECTING;

	uhttp_errno = 0;
	return 0;
}


int uhttp_prepare_filed_request(unsigned char *to_send, int to_send_size, int fd) {
	int ret;

	ret = uhttp_prepare_request(to_send, to_send_size, NULL, NULL);
	if (ret != 0)
		return ret;
	is_filep_request = 0;
	is_filed_request = 1;
	is_file_request = 1;
	filed = fd;
	file_size = 0;
	remain_file_size = 0;
	return 0;
}


int uhttp_prepare_filep_request(unsigned char *to_send, int to_send_size, FILE *fp) {
	int ret;

	ret = uhttp_prepare_request(to_send, to_send_size, NULL, NULL);
	if (ret != 0)
		return ret;
	is_filep_request = 1;
	is_filed_request = 0;
	is_file_request = 1;
	filep = fp;
	file_size = 0;
	remain_file_size = 0;
	return 0;
}


/*
static int check_socket(void) {
	int err;
	socklen_t err_size;
	char *errstr;

	if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &err, &err_size) == -1) {
		perror("getsockopt");
		return 1;
	}
	errstr = strerror(err);
	printf("check_socket: '%s'\n", errstr);
	return 0;
}
*/


/*
 * errors: UHTTP_ERROR_UNABLE_TO_CONNECT UHTTP_ERROR_UNABLE_TO_SEND UHTTP_ERROR_UNABLE_TO_RECEIVE UHTTP_TRY_AGAIN UHTTP_ERROR_SOCKET
 */
int uhttp_poll_request(long sec, long usec) {
	struct timeval tv;

	if ((uhttp_status != UHTTP_STATUS_SENDING) && (uhttp_status != UHTTP_STATUS_RECEIVING) && (uhttp_status != UHTTP_STATUS_CONNECTING))
		return 1;
	uhttp_errno = 0;
	//printf("uhttp_poll_request\n");
	if (uhttp_status == UHTTP_STATUS_CONNECTING) {
		//printf("\tuhttp_poll_request: connecting\n");
		remote.sin_family = AF_INET;
		remote.sin_port = htons(port);
		remote.sin_addr = *((struct in_addr *) he->h_addr);
		memset(&(remote.sin_zero), 0, 8);
		if (connect(sockfd, (struct sockaddr *) &remote, sizeof(struct sockaddr)) == -1) {
			uhttp_errno = UHTTP_ERROR_UNABLE_TO_CONNECT;
			return 1;
		}
		uhttp_status = UHTTP_STATUS_SENDING;
		uhttp_transfer_speed_init();
	}
	if (uhttp_status == UHTTP_STATUS_SENDING) {
		fd_set write_set, except_set;
		//printf("\tuhttp_poll_request: sending (i = %d)\n", to_send_buffer_i);
		while (1) {
			FD_ZERO(&write_set);
			FD_SET(sockfd, &write_set);
			FD_ZERO(&except_set);
			FD_SET(sockfd, &except_set);
			tv.tv_sec = sec;
			tv.tv_usec = usec;
			if (select(sockfd + 1, NULL, &write_set, &except_set, &tv) != 1) {
				uhttp_errno = UHTTP_ERROR_UNABLE_TO_SEND;
				return 1;
			}
			if (FD_ISSET(sockfd, &except_set)) {
				uhttp_errno = UHTTP_ERROR_SOCKET;
				close(sockfd);
				sockfd = -1;
				uhttp_status = UHTTP_STATUS_IDLE;
				return 1;
			}
			if (!FD_ISSET(sockfd, &write_set)) {
				uhttp_errno = UHTTP_TRY_AGAIN;
				return 1;
			}
			write(sockfd, to_send_buffer_p, 1);
			//printf("%c", *to_send_buffer_p);
			uhttp_transfer_speed_work(1);
			if (uhttp_sending_callback != NULL)
				uhttp_sending_callback(1);
			//send(sockfd, to_send_buffer_p, 1, 0);
			to_send_buffer_p++;
			to_send_buffer_i++;
			if (to_send_buffer_i >= to_send_buffer_size) {
				//printf("\tuhttp_poll_request: sending ok, now receiving\n");
				uhttp_status = UHTTP_STATUS_RECEIVING;
				uhttp_send_transfer_speed = uhttp_transfer_speed;
				uhttp_transfer_speed_init();
				if ((to_send_buffer != aux_buffer) && (to_send_buffer != NULL)) {
					free(to_send_buffer);
					to_send_buffer = NULL;
				}
				reading_header = 1;
				header_buffer_p = header_buffer;
				if (!is_file_request) {
					*to_receive_buffer = NULL;
					to_receive_buffer_p = NULL;
					to_receive_buffer_i = 0;
					*to_receive_buffer_size = 0;
				}
				break;
			}
		}
	}
	if (uhttp_status == UHTTP_STATUS_RECEIVING) {
		fd_set read_set, except_set;
		//printf("\tuhttp_poll_request: receiving\n");
		while (1) {
			FD_ZERO(&read_set);
			FD_SET(sockfd, &read_set);
			FD_ZERO(&except_set);
			FD_SET(sockfd, &except_set);
			tv.tv_sec = sec;
			tv.tv_usec = usec;
			if (select(sockfd + 1, &read_set, NULL, &except_set, &tv) != 1) {
				uhttp_errno = UHTTP_ERROR_UNABLE_TO_RECEIVE;
				return 1;
			}
			if (FD_ISSET(sockfd, &except_set)) {
				uhttp_errno = UHTTP_ERROR_SOCKET;
				close(sockfd);
				sockfd = -1;
				uhttp_status = UHTTP_STATUS_IDLE;
				return 1;
			}
			if (!FD_ISSET(sockfd, &read_set)) {
				uhttp_errno = UHTTP_TRY_AGAIN;
				return 1;
			}
			if (reading_header) {
				read(sockfd, header_buffer_p, 1);
				// recv(sockfd, header_buffer_p, 1, 0);
				//printf("%c", *header_buffer_p);
				uhttp_transfer_speed_work(1);
				if (uhttp_receiving_callback != NULL)
					uhttp_receiving_callback(1);
				//recv(sockfd, header_buffer_p, 1, 0);
				if (*header_buffer_p == '\r')
					continue;
				if (*header_buffer_p == '\n') {
					*header_buffer_p = 0;
					//printf("'%s'\n", header_buffer);
					if (header_buffer[0] == 0)
						reading_header = 0;
					else {
						if (sscanf((char *) header_buffer, "%[^:]: %s", parameter_name, parameter_value) == 2) {
							//printf("'%s' = '%s'\n", parameter_name, parameter_value);
							if (strcasecmp(parameter_name, "Set-Cookie") == 0)
								strcpy(cookie, parameter_value);
							if (strcasecmp(parameter_name, "Content-Length") == 0) {
								if (is_file_request) {
									sscanf(parameter_value, "%d", &file_size);
									remain_file_size = file_size;
								}
								else {
									sscanf(parameter_value, "%d", to_receive_buffer_size);
									printf("size = %d\n", *to_receive_buffer_size);
									if ((*to_receive_buffer = (unsigned char *) malloc(*to_receive_buffer_size)) == NULL) {
										uhttp_errno = UHTTP_ERROR_NO_MEM;
										return 1;
									}
									to_receive_buffer_p = *to_receive_buffer;
								}
							}
						}
					}
					header_buffer_p = header_buffer;
				}
				else
					header_buffer_p++;
			}
			else {
				int readed;
				/* reading data */
				readed = read(sockfd, uhttp_data, CHUNK_SIZE);
				//readed = recv(sockfd, uhttp_data, CHUNK_SIZE, 0);
				/*
				for (i = 0; i < readed; i++)
					printf("%c", uhttp_data[i]);
				*/
				uhttp_transfer_speed_work(readed);
				if (uhttp_receiving_callback != NULL)
					uhttp_receiving_callback(readed);
				//check_socket();
				//readed = recv(sockfd, uhttp_data, CHUNK_SIZE, 0);
				if (is_file_request) {
					if (readed > 0) {
						int to_passthrow;
						if (file_size >= 0) {
							if ((remain_file_size - readed) >= 0)
								to_passthrow = readed;
							else
								to_passthrow = remain_file_size;
							remain_file_size -= to_passthrow;
						}
						else
							to_passthrow = readed;
						if (is_filed_request) {
							//printf("write %d bytes\n", to_passthrow);
							write(filed, uhttp_data, to_passthrow);
						}
						else if (is_filep_request)
							fwrite(uhttp_data, 1, to_passthrow, filep);
						if ((remain_file_size == 0) && (file_size > 0)) {
							uhttp_recv_transfer_speed = uhttp_transfer_speed;
							close(sockfd);
							sockfd = -1;
							uhttp_status = UHTTP_STATUS_IDLE;
							uhttp_errno = 0;
							return 0;
						}
					}
					else {
						/*
						if (readed < 0)
							printf("ERROR!!!\n");
						*/
						uhttp_recv_transfer_speed = uhttp_transfer_speed;
						close(sockfd);
						sockfd = -1;
						uhttp_status = UHTTP_STATUS_IDLE;
						uhttp_errno = 0;
						return 0;
					}
				}
				else if (*to_receive_buffer != NULL) {
					if (readed > 0) {
						//*to_receive_buffer_p = uhttp_data;
						memcpy(to_receive_buffer_p, uhttp_data, readed);
						to_receive_buffer_p += readed;
						to_receive_buffer_i += readed;
						if (to_receive_buffer_i >= *to_receive_buffer_size) {
							uhttp_recv_transfer_speed = uhttp_transfer_speed;
							//printf("aqui\n");
							close(sockfd);
							sockfd = -1;
							uhttp_status = UHTTP_STATUS_IDLE;
							uhttp_errno = 0;
							return 0;
						}
					}
					else {
						uhttp_recv_transfer_speed = uhttp_transfer_speed;
						close(sockfd);
						sockfd = -1;
						uhttp_status = UHTTP_STATUS_IDLE;
						uhttp_errno = 0;
						return 0;
					}
				}
			}
		}
	}
	return 1;
}

/*
int main(int argc, char **argv) {
	char ts[] = "hola";
	char *tr;
	int tr_size;
	int s;

	s = uhttp_connect("http://panza:99/cgi-bin/test2.sh");

	
	printf("test 1\n");
	uhttp_prepare_request(ts, strlen(ts), (unsigned char **) &tr, &tr_size);
	while (uhttp_poll_request(0, 0) != 0)
		;
	tr[tr_size] = 0;
	printf("'%s'\n", tr);
	free(tr);

	
	printf("test 2\n");
	uhttp_prepare_request(ts, strlen(ts), (unsigned char **) &tr, &tr_size);
	while (uhttp_poll_request(0, 0) != 0)
		;
	tr[tr_size] = 0;
	printf("'%s'\n", tr);
	free(tr);


	return 0;
}
*/
